import abc
from functools import cache, cached_property
from typing import Union

from overhave.factory import IAdminFactory, ITestExecutionFactory
from overhave.pytest_plugin.config_injector import ConfigInjector
from overhave.pytest_plugin.plugin_resolver import IPluginResolver, PluginResolver

AnyProxyFactory = Union[IAdminFactory | ITestExecutionFactory]


class IProxyManager(abc.ABC):
    """Abstract class for proxy manager."""

    @property
    @abc.abstractmethod
    def injector(self) -> ConfigInjector:
        pass

    @abc.abstractmethod
    def set_factory(self, factory: AnyProxyFactory) -> None:
        pass

    @property
    @abc.abstractmethod
    def factory(self) -> AnyProxyFactory:
        pass

    @property
    @abc.abstractmethod
    def has_factory(self) -> bool:
        pass

    @abc.abstractmethod
    def clear_factory(self) -> None:
        pass

    @abc.abstractmethod
    def patch_pytest(self) -> None:
        pass

    @property
    @abc.abstractmethod
    def plugin_resolver(self) -> IPluginResolver:
        pass

    @property
    @abc.abstractmethod
    def pytest_patched(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def collection_prepared(self) -> bool:
        pass

    @abc.abstractmethod
    def supply_injector_for_collection(self) -> None:
        pass


class BaseProxyManagerException(Exception):
    """Base exception for :class:`ProxyManager`."""


class FactoryAlreadyDefinedError(BaseProxyManagerException):
    """Exception for situation with already defined `factory`."""


class FactoryNotDefinedError(BaseProxyManagerException):
    """Exception for situation with not defined `factory`."""


class ProxyManager(IProxyManager):
    """Manager for application factory resolution and usage, based on proxy-object pattern."""

    def __init__(self) -> None:
        self._factory: AnyProxyFactory | None = None
        self._pytest_patched = False
        self._collection_prepared = False

    @cached_property
    def _injector(self) -> ConfigInjector:
        return ConfigInjector()

    @property
    def injector(self) -> ConfigInjector:
        return self._injector

    def set_factory(self, factory: AnyProxyFactory) -> None:
        if self._factory is not None:
            raise FactoryAlreadyDefinedError("Factory is not nullable!")
        self._factory = factory

    @property
    def factory(self) -> AnyProxyFactory:
        if self._factory is None:
            raise FactoryNotDefinedError("Factory is nullable!")
        return self._factory

    def clear_factory(self) -> None:
        self._factory = None

    @property
    def has_factory(self) -> bool:
        return self._factory is not None

    def patch_pytest(self) -> None:
        if not self._pytest_patched:
            self._injector.patch_pytestbdd_prefixes(
                custom_step_prefixes=self.factory.context.language_settings.step_prefixes
            )
            self._pytest_patched = True

    @cached_property
    def _plugin_resolver(self) -> IPluginResolver:
        return PluginResolver(self.factory.context.file_settings)

    @property
    def plugin_resolver(self) -> IPluginResolver:
        return self._plugin_resolver

    @property
    def pytest_patched(self) -> bool:
        return self._pytest_patched

    @property
    def collection_prepared(self) -> bool:
        return self._collection_prepared

    def supply_injector_for_collection(self) -> None:
        if not self._collection_prepared:
            self._injector.supplement_components(
                file_settings=self.factory.context.file_settings,
                step_collector=self.factory.step_collector,
                test_runner=self.factory.test_runner,
                feature_extractor=self.factory.feature_extractor,
            )
            self._collection_prepared = True


@cache
def get_proxy_manager() -> IProxyManager:
    return ProxyManager()



import logging
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import cast

import allure
import pytz

from overhave import db
from overhave.entities import BaseFileExtractor, FeatureExtractor, GitRepositoryInitializer, OverhaveFileSettings
from overhave.scenario import FeatureInfo, NullableFeatureIdError, ScenarioParser, StrictFeatureInfo
from overhave.storage import FeatureModel
from overhave.synchronization.abstract import IOverhaveSynchronizer
from overhave.synchronization.storage_manager import SynchronizerStorageManager
from overhave.utils import ANY_INT, get_current_time

logger = logging.getLogger(__name__)


class BaseOverhaveSynchronizerException(Exception):
    """Base exception for :class:`OverhaveSynchronizer`."""


class NullableInfoNameError(BaseOverhaveSynchronizerException):
    """Exception for situation without feature info name."""


class NullableInfoAuthorError(BaseOverhaveSynchronizerException):
    """Exception for situation without feature info author."""


class NullableInfoFeatureTypeError(BaseOverhaveSynchronizerException):
    """Exception for situation without feature info type."""


class OverhaveSynchronizer(BaseFileExtractor, IOverhaveSynchronizer):
    """Class for synchronization between git and database."""

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        scenario_parser: ScenarioParser,
        feature_extractor: FeatureExtractor,
        git_initializer: GitRepositoryInitializer,
        storage_manager: SynchronizerStorageManager,
    ):
        super().__init__(extenstion=file_settings.feature_suffix)
        self._file_settings = file_settings
        self._scenario_parser = scenario_parser
        self._feature_extractor = feature_extractor
        self._git_initializer = git_initializer
        self._storage_manager = storage_manager

    def _update_feature(self, model: FeatureModel, info: StrictFeatureInfo, file_ts: datetime) -> None:
        logger.info("Feature is gonna be updated...")
        with db.create_session() as session:
            self._storage_manager.ensure_users_exist(session=session, info=info)
            model.feature_tags = self._storage_manager.get_feature_tags(session=session, info=info)
        model.name = info.name
        model.severity = info.severity
        model.last_edited_by = info.last_edited_by
        model.last_edited_at = file_ts
        model.task = info.tasks
        self._storage_manager.update_db_feature(model=model, scenario=info.scenarios)
        logger.info("Feature has been updated successfully.")

    def _create_feature(self, file: Path, info: FeatureInfo) -> None:
        logger.info("Feature is gonna be created...")
        if info.name is None:
            raise NullableInfoNameError("Feature info has not got feature name!")
        if info.type is None:
            raise NullableInfoFeatureTypeError("Could not create feature without feature type!")
        if info.author is None:
            raise NullableInfoAuthorError("Could not create feature without author!")
        if info.last_edited_by is None:
            info.last_edited_by = info.author

        info.id = ANY_INT
        strict_info = StrictFeatureInfo(**asdict(info))

        with db.create_session() as session:
            self._storage_manager.ensure_users_exist(session=session, info=strict_info)
            feature_tags = self._storage_manager.get_feature_tags(session=session, info=strict_info)
            feature_type = self._storage_manager.feature_type_by_name(session=session, feature_type=strict_info.type)

        feature_model = FeatureModel(
            id=strict_info.id,
            created_at=get_current_time(),
            name=strict_info.name,
            author=strict_info.author,
            type_id=feature_type.id,
            last_edited_by=strict_info.last_edited_by,
            last_edited_at=strict_info.last_edited_at,
            task=strict_info.tasks or [],
            file_path=file.relative_to(
                self._feature_extractor.feature_type_to_dir_mapping[feature_type.name]
            ).as_posix(),
            released=True,
            feature_type=feature_type,
            feature_tags=feature_tags,
            severity=strict_info.severity or allure.severity_level.NORMAL,
        )
        self._storage_manager.create_db_feature(model=feature_model, scenario=strict_info.scenarios)
        logger.info("Feature with ID=%s has been created successfully.", feature_model.id)

    def synchronize(self, create_db_features: bool = False, pull_repository: bool = False) -> None:  # noqa: C901
        if pull_repository:
            self._git_initializer.pull()
        logger.info("Start synchronization...")
        all_features = self._extract_recursively(self._file_settings.features_dir)
        for feature_file in all_features:
            logger.info("Synchronize feature from file %s...", feature_file.as_posix())
            feature_text = feature_file.read_text()
            self._scenario_parser.set_strict_mode(True)
            try:
                feature_info: StrictFeatureInfo = cast(StrictFeatureInfo, self._scenario_parser.parse(feature_text))
            except NullableFeatureIdError:
                logger.warning("Feature has not got Overhave ID or ID format is incorrect.")
                if not create_db_features:
                    logger.warning("create_db_features=%s. Skip it.", create_db_features)
                    continue
                self._scenario_parser.set_strict_mode(False)
                optional_feature_info = cast(FeatureInfo, self._scenario_parser.parse(feature_text))
                self._create_feature(file=feature_file, info=optional_feature_info)
                continue
            logger.debug("Parsed feature info: %s", feature_info)
            feature_model = self._storage_manager.get_feature(feature_info.id)
            if feature_model is None:
                logger.warning("Feature doesn't exist in Overhave database.")
                continue  # TODO: unlink file and create MR with deletions at the end

            feature_file_ts = datetime.fromtimestamp(feature_file.stat().st_mtime, tz=pytz.UTC)
            if feature_model.last_edited_at == feature_file_ts and feature_model.released:
                logger.warning("Feature has been already synchronized.")
                continue
            last_change_time = self._storage_manager.get_last_change_time(model=feature_model)
            if last_change_time < feature_file_ts:
                self._update_feature(model=feature_model, info=feature_info, file_ts=feature_file_ts)
                continue
            if feature_model.released:
                logger.info("Feature is already actual. Skip.")
                continue
            logger.warning(
                "Feature was changed soon (at %s), but not released. Skip.",
                last_change_time.strftime("%d-%m-%Y %H:%M:%S"),
            )
        logger.info("Synchronization completed.")



import httpx
from pydantic import Field, field_validator

from overhave.base_settings import BaseOverhavePrefix
from overhave.utils import make_url


class EmptyOverhaveAdminURLError(ValueError):
    """Exception for situation with empty ````admin_url``` while trying to ```get_feature_url```."""


class OverhaveAdminLinkSettings(BaseOverhavePrefix):
    """Settings for dynamic links to Overhave Admin in Allure report."""

    admin_url: httpx.URL | None

    feature_id_filter_path: str = "feature/?flt2_0={feature_id}"
    feature_id_placeholder: str = "Overhave feature #{feature_id}"

    @field_validator("admin_url", mode="before")
    def make_admin_url(cls, v: str | None) -> httpx.URL | None:
        return make_url(v)

    @property
    def enabled(self) -> bool:
        return self.admin_url is not None

    def get_feature_url(self, feature_id: int) -> httpx.URL:
        if isinstance(self.admin_url, httpx.URL):
            return httpx.URL(f"{self.admin_url}/{self.feature_id_filter_path.format(feature_id=feature_id)}")
        raise EmptyOverhaveAdminURLError("Overhave admin URL is None, so could not create link URL!")

    def get_feature_link_name(self, feature_id: int) -> str:
        return self.feature_id_placeholder.format(feature_id=feature_id)


class OverhaveTestSettings(BaseOverhavePrefix):
    """Settings for PytestRunner, which runs scenario tests with specified addoptions."""

    default_pytest_addoptions: str = "--disable-warnings"
    extra_pytest_addoptions: str | None = Field(default=None)

    workers: int | None = Field(default=None, description="Number of xdist workers")


class OverhaveStepCollectorSettings(BaseOverhavePrefix):
    """Settings for StepCollector, which collect BDD steps for Overhave Admin UI."""

    hide_non_public_steps: bool = False



from http import HTTPStatus

import fastapi
from fastapi import security as fastapi_security
from jose import JWTError

from overhave import db
from overhave.api.auth.models import AUTH_HEADERS
from overhave.api.auth.token import get_token_data
from overhave.api.deps import get_api_auth_settings, get_system_user_storage
from overhave.api.settings import OverhaveApiAuthSettings
from overhave.storage import ISystemUserStorage, SystemUserModel

oauth2_scheme = fastapi_security.OAuth2PasswordBearer(tokenUrl="token")


def get_authorized_user(
    token: str = fastapi.Depends(oauth2_scheme),
    auth_settings: OverhaveApiAuthSettings = fastapi.Depends(get_api_auth_settings),
    storage: ISystemUserStorage = fastapi.Depends(get_system_user_storage),
) -> SystemUserModel:
    creds_exception = fastapi.HTTPException(
        status_code=HTTPStatus.UNAUTHORIZED,
        detail="Could not validate credentials",
        headers=AUTH_HEADERS.model_dump(),
    )
    try:
        token_data = get_token_data(auth_settings=auth_settings, token=token)
        if token_data is None:
            raise creds_exception
    except JWTError:
        raise creds_exception
    with db.create_session() as session:
        user = storage.get_user_by_credits(session=session, login=token_data.username)
        if user is None:
            raise creds_exception
        return SystemUserModel.model_validate(user)



from pathlib import Path
from typing import cast

from _pytest.nodes import Item
from pytest_bdd.parser import Scenario

from overhave.pytest_plugin.helpers.bdd_item import get_scenario
from overhave.scenario import FeatureInfo, ScenarioParser


def _parse_feature_info_from_file(scenario: Scenario, scenario_parser: ScenarioParser) -> FeatureInfo:
    feature_txt = Path(scenario.feature.filename).read_text()
    return cast(FeatureInfo, scenario_parser.parse(feature_txt))


def set_feature_info_for_item(item: Item, scenario_parser: ScenarioParser) -> None:
    setattr(
        item,
        "feature_info",
        _parse_feature_info_from_file(scenario=get_scenario(item), scenario_parser=scenario_parser),
    )


def get_feature_info_from_item(item: Item) -> FeatureInfo:
    feature_info: FeatureInfo | None = getattr(item, "feature_info")
    if feature_info is None:
        raise AttributeError(f"Item {item} has not got attr 'feature_info'!")
    return feature_info



# flake8: noqa

from .client import (
    BaseOverhaveMetricContainer,
    EmulationRunOverhaveMetricContainer,
    PublicationOverhaveMetricContainer,
    TestRunOverhaveMetricContainer,
)
from .getters import (
    get_common_metric_container,
    get_emulation_metric_container,
    get_publication_metric_container,
    get_test_metric_container,
)



import logging
from http import HTTPStatus

import httpx
from gitlab import GitlabError

from overhave import db
from overhave.entities import GitRepositoryInitializer
from overhave.metrics import PublicationOverhaveMetricContainer
from overhave.publication.git_publisher import GitVersionPublisher
from overhave.publication.gitlab.settings import OverhaveGitlabPublisherSettings
from overhave.publication.gitlab.tokenizer.client import TokenizerClient
from overhave.scenario import FileManager, OverhaveProjectSettings
from overhave.storage import IDraftStorage, IFeatureStorage, IScenarioStorage, ITestRunStorage
from overhave.transport.http.gitlab_client import GitlabHttpClient, GitlabMrRequest

logger = logging.getLogger(__name__)


class GitlabVersionPublisher(GitVersionPublisher[OverhaveGitlabPublisherSettings]):
    """Class for feature version's merge requests management relative to Gitlab API."""

    def __init__(
        self,
        project_settings: OverhaveProjectSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        draft_storage: IDraftStorage,
        file_manager: FileManager,
        git_initializer: GitRepositoryInitializer,
        gitlab_publisher_settings: OverhaveGitlabPublisherSettings,
        gitlab_client: GitlabHttpClient,
        tokenizer_client: TokenizerClient,
        metric_container: PublicationOverhaveMetricContainer,
    ):
        super().__init__(
            project_settings=project_settings,
            feature_storage=feature_storage,
            scenario_storage=scenario_storage,
            test_run_storage=test_run_storage,
            draft_storage=draft_storage,
            file_manager=file_manager,
            git_initializer=git_initializer,
            git_publisher_settings=gitlab_publisher_settings,
            metric_container=metric_container,
        )
        self._gitlab_client = gitlab_client
        self._tokenizer_client = tokenizer_client

    def publish_version(self, draft_id: int) -> db.DraftStatus:
        context = self._prepare_publisher_context(draft_id)
        if context is None:
            self._draft_storage.set_draft_status(draft_id=draft_id, status=db.DraftStatus.INTERNAL_ERROR)
            return db.DraftStatus.INTERNAL_ERROR
        merge_request = GitlabMrRequest(
            project_id=self._git_publisher_settings.repository_id,
            title=context.feature.name,
            source_branch=context.target_branch,
            target_branch=self._git_publisher_settings.target_branch,
            description=self._compile_publication_description(context),
            reviewer_ids=self._git_publisher_settings.get_reviewers(feature_type=context.feature.feature_type.name),
        )
        logger.info("Prepared merge-request: %s", merge_request.model_dump_json(by_alias=True))
        try:
            token = (
                self._gitlab_client._settings.auth_token or self._tokenizer_client.get_token(draft_id=draft_id).token
            )
            response = self._gitlab_client.send_merge_request(
                merge_request=merge_request, token=token, repository_id=self._git_publisher_settings.repository_id
            )
            self._draft_storage.save_response_as_created(
                draft_id=draft_id,
                pr_url=response.web_url,
                published_at=response.created_at,
            )
            logger.info("Draft.id=%s successfully sent to GitLab", draft_id)
            return db.DraftStatus.CREATED
        except (GitlabError, httpx.HTTPError) as err:
            logger.exception("Got error while trying to sent merge-request!")
            if isinstance(err, GitlabError) and err.response_code == HTTPStatus.CONFLICT:
                self._draft_storage.save_response_as_duplicate(
                    draft_id=context.draft.id, feature_id=context.feature.id, traceback=str(err)
                )
                return db.DraftStatus.DUPLICATE
            self._draft_storage.set_draft_status(
                draft_id=draft_id, status=db.DraftStatus.INTERNAL_ERROR, traceback=str(err)
            )
            return db.DraftStatus.INTERNAL_ERROR



# flake8: noqa
from .api_auth_storage import AuthStorage, AuthToken, IAuthStorage
from .converters import (
    DraftModel,
    EmulationModel,
    EmulationRunModel,
    FeatureModel,
    FeatureTypeModel,
    FeatureTypeName,
    PublisherContext,
    ScenarioModel,
    SystemUserModel,
    TagModel,
    TestExecutorContext,
    TestRunModel,
    TestUserModel,
    TestUserSpecification,
)
from .draft_storage import DraftStorage, IDraftStorage, NullableScenarioError
from .emulation_storage import EmulationStorage, EmulationStorageError, IEmulationStorage
from .feature_storage import FeatureStorage, IFeatureStorage
from .feature_tag_storage import FeatureTagStorage, IFeatureTagStorage
from .feature_type_storage import FeatureTypeNotExistsError, FeatureTypeStorage, IFeatureTypeStorage
from .scenario_storage import IScenarioStorage, ScenarioStorage
from .system_user_group_storage import ISystemUserGroupStorage, SystemUserGroupStorage
from .system_user_storage import ISystemUserStorage, SystemUserStorage
from .test_run_storage import ITestRunStorage, TestRunStorage
from .test_user_storage import (
    ITestUserStorage,
    TestUserDoesNotExistError,
    TestUserStorage,
    TestUserUpdatingNotAllowedError,
)



import enum
import logging
import socket
from logging.config import DictConfigurator
from typing import Any

import sqlalchemy as sa
import sqlalchemy.exc
from pydantic import field_validator
from pydantic_settings import BaseSettings
from sqlalchemy.pool import Pool, SingletonThreadPool

OVERHAVE_ENV_PREFIX = "OVERHAVE_"


class BaseOverhavePrefix(BaseSettings):
    """Possibility to change Overhave default settings from environment."""

    class Config:
        env_prefix = OVERHAVE_ENV_PREFIX


def _as_alchemy_url(db_url: str) -> sa.URL:
    try:
        return sa.make_url(db_url)
    except sqlalchemy.exc.ArgumentError as e:
        raise ValueError from e


class DataBaseSettings(BaseOverhavePrefix):
    """Overhave database settings."""

    db_url: str = "postgresql://postgres:postgres@localhost/overhave"
    db_pool_recycle: int = 500
    db_pool_size: int = 6
    db_echo: bool = False
    db_application_name: str = socket.gethostname()
    db_connect_timeout: int = 30
    db_poolclass: type[Pool] = SingletonThreadPool

    def _create_engine(self) -> sa.Engine:
        return sa.engine_from_config(
            {
                "url": _as_alchemy_url(self.db_url),
                "pool_recycle": self.db_pool_recycle,
                "pool_pre_ping": True,
                "pool_size": self.db_pool_size,
                "poolclass": self.db_poolclass,
                "connect_args": {
                    "connect_timeout": self.db_connect_timeout,
                    "application_name": self.db_application_name,
                },
            },
            prefix="",
        )

    def setup_engine(self) -> None:
        from overhave.db.base import current_session, metadata

        metadata.set_engine(engine=self._create_engine())
        current_session.configure(bind=metadata.engine)


class LoggingSettings(BaseOverhavePrefix):
    """Overhave logging settings."""

    log_level: str = logging.getLevelName(logging.INFO)
    log_config: dict[str, Any] = {}

    @field_validator("log_config", check_fields=True)
    def dict_config_validator(cls, v: dict[str, Any]) -> DictConfigurator | None:
        if not v:
            return None
        return DictConfigurator(v)

    def setup_logging(self) -> None:
        if isinstance(self.log_config, DictConfigurator):
            self.log_config.configure()
        else:
            logging.basicConfig(level=self.log_level)


class AuthorizationStrategy(enum.StrEnum):
    """
    Authorization strategies Enum.

    Simple - strategy without real auth_managers. Each user could use preferred name. This name will be used for user
    authority. Each user is unique. Password not required.
    Default - strategy with real auth_managers. Each user could use only registered credentials.
    LDAP - strategy with auth_managers using remote LDAP server. Each user should use his LDAP credentials. LDAP
    server returns user groups. If user in default 'admin' group or his groups list contains admin group - user
    will be authorized. If user already placed in database - user will be authorized too. No one password stores.
    """

    SIMPLE = "simple"
    DEFAULT = "default"
    LDAP = "ldap"


class OverhaveAuthorizationSettings(BaseOverhavePrefix):
    """Settings for Overhave auth_managers in components interface.

    Supports 3 strategies: SIMPLE, DEFAULT and LDAP.
    LDAP auth_managers uses group politics with administration group `admin_group`.
    SIMPLE and DEFAULT strategies use admin user that would be dynamically created at startup.
    """

    auth_strategy: AuthorizationStrategy = AuthorizationStrategy.SIMPLE



from datetime import datetime

from pydantic import BaseModel, Field


class GitlabRepository(BaseModel):
    """Model for Gitlab merge-request repository."""

    id: str = Field(..., alias="project_id")


class GitlabMrRequest(BaseModel):
    """Model for Gitlab merge-request request."""

    id: str = Field(..., alias="project_id")
    source_branch: str
    target_branch: str
    title: str | None
    description: str | None
    reviewer_ids: list[str]


class GitlabMrCreationResponse(BaseModel):
    """Model for Gitlab merge-request creation response."""

    created_at: datetime
    web_url: str



import abc
import logging
import socket
from typing import cast

import sqlalchemy as sa
import sqlalchemy.orm as so

from overhave import db
from overhave.entities.settings import OverhaveEmulationSettings
from overhave.storage import EmulationRunModel
from overhave.utils import get_current_time

logger = logging.getLogger(__name__)


class EmulationStorageError(Exception):
    """Base Exception for :class:`EmulationStorage`."""


class AllPortsAreBusyError(EmulationStorageError):
    """Exception for situation when all ports are busy."""


class IEmulationStorage(abc.ABC):
    """Abstract class for emulation runs storage."""

    @staticmethod
    @abc.abstractmethod
    def create_emulation_run(emulation_id: int, initiated_by: str) -> int:
        pass

    @abc.abstractmethod
    def get_requested_emulation_run(self, emulation_run_id: int) -> EmulationRunModel:
        pass

    @abc.abstractmethod
    def set_emulation_run_status(self, emulation_run_id: int, status: db.EmulationStatus) -> None:
        pass

    @abc.abstractmethod
    def set_error_emulation_run(self, emulation_run_id: int, traceback: str) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_emulation_runs_by_test_user_id(test_user_id: int) -> list[EmulationRunModel]:
        pass


class EmulationStorage(IEmulationStorage):
    """Class for emulation runs storage."""

    def __init__(self, settings: OverhaveEmulationSettings):
        self._settings = settings
        self._emulation_ports_len = len(self._settings.emulation_ports)

    @staticmethod
    def create_emulation_run(emulation_id: int, initiated_by: str) -> int:
        with db.create_session() as session:
            emulation_run = db.EmulationRun(
                emulation_id=emulation_id, initiated_by=initiated_by, status=db.EmulationStatus.CREATED
            )
            session.add(emulation_run)
            session.flush()
            return emulation_run.id

    def _get_next_port(self, session: so.Session) -> int:
        runs_with_allocated_ports = (  # noqa: ECE001
            session.query(db.EmulationRun)
            .filter(db.EmulationRun.port.isnot(None))
            .order_by(db.EmulationRun.id.desc())
            .limit(self._emulation_ports_len)
            .all()
        )
        allocated_sorted_runs = sorted(
            runs_with_allocated_ports,
            key=lambda t: t.changed_at,
        )

        allocated_ports = {run.port for run in allocated_sorted_runs}
        logger.debug("Allocated ports: %s", allocated_ports)
        not_allocated_ports = set(self._settings.emulation_ports).difference(allocated_ports)
        logger.debug("Not allocated ports: %s", not_allocated_ports)
        if not_allocated_ports:
            for port in not_allocated_ports:
                if self._is_port_in_use(port):
                    continue
                return port
            logger.debug("All not allocated ports are busy!")
        for run in allocated_sorted_runs:
            if self._is_port_in_use(cast(int, run.port)):
                continue
            return cast(int, run.port)
        raise AllPortsAreBusyError("All ports are busy - could not find free port!")

    def _is_port_in_use(self, port: int) -> bool:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex((self._settings.emulation_bind_ip, port)) == 0

    def get_requested_emulation_run(self, emulation_run_id: int) -> EmulationRunModel:
        with db.create_session() as session:
            emulation_run = session.query(db.EmulationRun).filter(db.EmulationRun.id == emulation_run_id).one()
            emulation_run.status = db.EmulationStatus.REQUESTED
            emulation_run.port = self._get_next_port(session)
            emulation_run.changed_at = get_current_time()
            return EmulationRunModel.model_validate(emulation_run)

    def set_emulation_run_status(self, emulation_run_id: int, status: db.EmulationStatus) -> None:
        with db.create_session() as session:
            session.execute(
                sa.update(db.EmulationRun)
                .where(db.EmulationRun.id == emulation_run_id)
                .values(status=status, changed_at=get_current_time())
            )

    def set_error_emulation_run(self, emulation_run_id: int, traceback: str) -> None:
        with db.create_session() as session:
            session.execute(
                sa.update(db.EmulationRun)
                .where(db.EmulationRun.id == emulation_run_id)
                .values(status=db.EmulationStatus.ERROR, changed_at=get_current_time(), traceback=traceback)
            )

    @staticmethod
    def get_emulation_runs_by_test_user_id(test_user_id: int) -> list[EmulationRunModel]:
        with db.create_session() as session:
            emulation_ids_query = (
                session.query(db.Emulation)
                .with_entities(db.Emulation.id)
                .filter(db.Emulation.test_user_id == test_user_id)
                .scalar_subquery()
            )
            emulation_runs = (
                session.query(db.EmulationRun).where(db.EmulationRun.emulation_id.in_(emulation_ids_query)).all()
            )
            return [EmulationRunModel.model_validate(x) for x in emulation_runs]



import logging
from pathlib import Path

from overhave.entities.feature.abstract import IFeatureExtractor
from overhave.entities.feature.errors import FeatureTypeExtractionError, ScenariosTestFileNotFound
from overhave.entities.settings import OverhaveFileSettings
from overhave.storage import FeatureTypeName

logger = logging.getLogger(__name__)


class FeatureExtractor(IFeatureExtractor):
    """Class for specified project's feature types resolution."""

    def __init__(self, file_settings: OverhaveFileSettings):
        self._file_settings = file_settings

        self._feature_types: list[FeatureTypeName] = []
        self._feature_type_to_dir_mapping: dict[FeatureTypeName, Path] = {}
        try:
            self._extract_project_data()
            self._check_pytest_bdd_scenarios_test_files()
        except FeatureTypeExtractionError:
            logger.exception("Extraction error while trying to collect features!")

    def _extract_project_data(self) -> None:
        try:
            feature_type_dirs = tuple(
                d
                for d in self._file_settings.features_dir.iterdir()
                if all(
                    (
                        d.is_dir(),
                        d != self._file_settings.tmp_dir,
                        not d.name.startswith("."),
                        not d.name.startswith("_"),
                    )
                )
            )
        except FileNotFoundError as err:
            raise FeatureTypeExtractionError(
                f"Could not find any subdirectory in specified features directory '{self._file_settings.features_dir}'!"
            ) from err
        self._feature_types = [FeatureTypeName(t.name) for t in feature_type_dirs]
        logger.info("Registered feature types: %s", self._feature_types)
        self._feature_type_to_dir_mapping = {FeatureTypeName(t.name): t for t in feature_type_dirs}

    def _check_pytest_bdd_scenarios_test_files(self) -> None:
        for feature_type in self._feature_types:
            scenarios_file = self._file_settings.fixtures_dir / self._file_settings.fixtures_file_template_mask.format(
                feature_type=feature_type
            )
            if scenarios_file.exists():
                continue
            raise ScenariosTestFileNotFound(
                "Could not find pytest-bdd test file with scenarios definition! "
                f"Maybe you don't created '{scenarios_file.name}' in '{scenarios_file.parent}' directory?"
            )

    @property
    def feature_types(self) -> list[FeatureTypeName]:
        return self._feature_types

    @property
    def feature_type_to_dir_mapping(self) -> dict[FeatureTypeName, Path]:
        return self._feature_type_to_dir_mapping



from typing import Generator

import httpx
import requests


class BearerAuth(httpx.Auth):
    """Class for Bearer auth_managers pattern."""

    def __init__(self, token: str) -> None:
        self.token = token

    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        request.headers["Authorization"] = f"Bearer {self.token}"
        yield request

    def __call__(self, request: httpx.Request | requests.PreparedRequest) -> httpx.Request | requests.PreparedRequest:
        return next(self.auth_flow(request=request))  # type: ignore



from overhave.entities import (
    OverhaveDescriptionManagerSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveReportManagerSettings,
    OverhaveStepContextSettings,
)
from overhave.factory.context.base_context import BaseFactoryContext
from overhave.scenario import OverhaveProjectSettings, OverhaveScenarioCompilerSettings, OverhaveScenarioParserSettings
from overhave.test_execution import OverhaveAdminLinkSettings, OverhaveTestSettings
from overhave.transport import OverhaveS3ManagerSettings


class OverhaveTestExecutionContext(BaseFactoryContext):
    """Overhave test run context, based on application BaseSettings.

    This context defines how Overhave test run will work.
    """

    def __init__(
        self,
        compilation_settings: OverhaveScenarioCompilerSettings | None = None,
        parser_settings: OverhaveScenarioParserSettings | None = None,
        description_manager_settings: OverhaveDescriptionManagerSettings | None = None,
        file_settings: OverhaveFileSettings | None = None,
        language_settings: OverhaveLanguageSettings | None = None,
        project_settings: OverhaveProjectSettings | None = None,
        report_manager_settings: OverhaveReportManagerSettings | None = None,
        s3_manager_settings: OverhaveS3ManagerSettings | None = None,
        step_context_settings: OverhaveStepContextSettings | None = None,
        test_settings: OverhaveTestSettings | None = None,
        admin_link_settings: OverhaveAdminLinkSettings | None = None,
    ) -> None:
        super().__init__(
            compilation_settings=compilation_settings or OverhaveScenarioCompilerSettings(),
            parser_settings=parser_settings or OverhaveScenarioParserSettings(),
            file_settings=file_settings or OverhaveFileSettings(),
            language_settings=language_settings or OverhaveLanguageSettings(),
            project_settings=project_settings or OverhaveProjectSettings(),
            report_manager_settings=report_manager_settings or OverhaveReportManagerSettings(),
            s3_manager_settings=s3_manager_settings or OverhaveS3ManagerSettings(),
            test_settings=test_settings or OverhaveTestSettings(),
        )
        self.description_manager_settings = description_manager_settings or OverhaveDescriptionManagerSettings()
        self.step_context_settings = step_context_settings or OverhaveStepContextSettings()
        self.admin_link_settings = admin_link_settings or OverhaveAdminLinkSettings()



import logging

from pydantic import SecretStr

from overhave import db
from overhave.entities.auth_managers.secret_mixin import AdminSecretMixin
from overhave.storage import SystemUserModel

logger = logging.getLogger(__name__)


class BaseSimpleAdminAuthorizationManagerException(Exception):
    """Base exception for :class:`SimpleAdminAuthorizationManager`."""


class NullablePasswordError(Exception):
    """Exception for situation when row is without password."""


class SimpleAdminAuthorizationManager(AdminSecretMixin):
    """Class for user registration.

    Manager does not provide real auth_managers. Each user could use preferred name.
    This name will be used for user authority. Each user is unique. Passwords not required.
    """

    def authorize_user(self, username: str, password: SecretStr) -> SystemUserModel | None:
        with db.create_session() as session:
            user = self._system_user_storage.get_user_by_credits(session=session, login=username)
            if user is None:
                user = self._system_user_storage.create_user(session=session, login=username, password=password)
            if user.password is None:
                raise NullablePasswordError(f"User with id={user.id} has not got password!")
            if user.password == password.get_secret_value():
                return SystemUserModel.model_validate(user)
        return None



import logging
from functools import cached_property
from pathlib import Path

from overhave.scenario.validator.errors import DuplicateFeatureIDError

logger = logging.getLogger(__name__)


class FeatureDuplicatedIdValidationMixin:
    """Class for features validation."""

    @cached_property
    def _feature_id_to_path_mapping(self) -> dict[int, list[Path]]:
        return {}

    def _save_to_feature_id_to_path_mapping(self, feature_path: Path, feature_id: int | None) -> None:
        if not isinstance(feature_id, int):
            return
        if not self._feature_id_to_path_mapping.get(feature_id):
            self._feature_id_to_path_mapping[feature_id] = [feature_path]
            return
        logger.warning("Found duplicate ID=%s!", feature_id)
        self._feature_id_to_path_mapping[feature_id].append(feature_path)

    def _validate_duplicate_ids(self) -> None:
        duplicate_feature_ids = [
            feature_id
            for feature_id, feature_paths in self._feature_id_to_path_mapping.items()
            if len(feature_paths) > 1
        ]
        if not duplicate_feature_ids:
            return
        txt = "Features with duplicate IDs:\n"
        raise DuplicateFeatureIDError(
            txt
            + "\n".join(
                "id={} - {}".format(x, [path.as_posix() for path in self._feature_id_to_path_mapping[x]])
                for x in duplicate_feature_ids
            )
        )



from overhave.entities import OverhaveFileSettings, OverhaveLanguageSettings
from overhave.factory.context.base_context import BaseFactoryContext
from overhave.scenario import OverhaveProjectSettings, OverhaveScenarioCompilerSettings, OverhaveScenarioParserSettings


class OverhaveSynchronizerContext(BaseFactoryContext):
    """Overhave feature synchronizer context, based on application BaseSettings.

    This context defines how Overhave admin will work.
    """

    def __init__(
        self,
        file_settings: OverhaveFileSettings | None = None,
        compilation_settings: OverhaveScenarioCompilerSettings | None = None,
        parser_settings: OverhaveScenarioParserSettings | None = None,
        language_settings: OverhaveLanguageSettings | None = None,
        project_settings: OverhaveProjectSettings | None = None,
    ) -> None:
        super().__init__(
            file_settings=file_settings or OverhaveFileSettings(),
            compilation_settings=compilation_settings or OverhaveScenarioCompilerSettings(),
            parser_settings=parser_settings or OverhaveScenarioParserSettings(),
            language_settings=language_settings or OverhaveLanguageSettings(),
            project_settings=project_settings or OverhaveProjectSettings(),
        )



import httpx

from overhave.base_settings import OVERHAVE_ENV_PREFIX
from overhave.transport.http.base_client import BaseHttpClientSettings


class OverhaveApiAuthenticatorSettings(BaseHttpClientSettings):
    """Settings for :class:`OverhaveApiAuthenticator`."""

    auth_token_path: str = "token"

    class Config:
        env_prefix = OVERHAVE_ENV_PREFIX + "API_AUTH_"

    @property
    def get_auth_token_url(self) -> httpx.URL:
        return httpx.URL(f"{self.url}/{self.auth_token_path}")



import abc

from overhave import db
from overhave.transport import PublicationTask


class IVersionPublisher(abc.ABC):
    """Abstract class for feature version's pull requests management."""

    @abc.abstractmethod
    def publish_version(self, draft_id: int) -> db.DraftStatus:
        pass

    @abc.abstractmethod
    def process_publish_task(self, task: PublicationTask) -> None:
        pass



from datetime import timedelta

from pydantic_settings import BaseSettings


class OverhaveLdapClientSettings(BaseSettings):
    """Settings for Overhave LDAP client."""

    url: str  # for example: "ldap://mydomain.ru"
    domain: str  # for example: "domain\\"
    dn: str  # for example: "dc=example,dc=com"
    timeout: timedelta = timedelta(seconds=10)
    tls_enabled: bool = True

    class Config:
        env_prefix = "OVERHAVE_LDAP_"
